/* PLPRINT.C : Printer output routines for PICLAB.  The only command
** handler here is for the PRINT command.
*/

#include <stdlib.h>
#include <string.h>
#include <memory.h>

#include "piclab.h"
#include "pjtbls"

static int outfile;
static char *titles[10];
static U8 *inlines[3], *outbuf;
static int ntitles, ptype, pwidth = 0, transparent = 0;

extern int sprintf(char *, const char *, ...);

static int startprint(char *);
static int endprint(void);
static int pjcolor(U8 *);
static int pjprint(void);
static int ljprint(void);

/* Open printer output file, write initialization string.
*/
static int startprint(char *filename)
{
    char initstring[80];
    int l;

    if ((outfile = p_open(filename, outbuf, BUFSIZE, 512, WRITE)) < 0) return 3;
    switch (ptype) {
        case 1: /* PAINTJET */
            sprintf(initstring,
            "\r\033&k%dW\033*t180R\033*r%ds3U\033*b%dM\033&a%dH\033*r1A",
            transparent ? 3 : 0, pwidth<<1, 0, 0);
            break;
        case 2: /* LASERJET */
            sprintf(initstring,
            "\r\033E\033&l0e0O\033&s1C\0339\033*t%dR\033*r1A",
            laserdpi);
            break;
        default:    return 1;
    }
    l = strlen(initstring);
    memcpy(linebuf(outfile), initstring, l);
    if (p_putline(outfile, l) != l) return 3;
    else return 0;
}

/* Write cleanup string, close output file.
*/
static int endprint()
{
    int i, l;
    char temp[80];

    switch (ptype) {
        case 1: /* PAINTJET */
        case 2: /* LASERJET */
            sprintf(temp, "\033*r0B\n");
            break;
        default:    return 1;
    }
    l = strlen(temp);
    memcpy(linebuf(outfile), temp, l);
    if (p_putline(outfile, l) != l) return 3;
    for (i=0; i<ntitles; ++i) {
        sprintf(temp, "%s ", titles[i]);
        l = strlen(temp);
        memcpy(linebuf(outfile), temp, l);
        if (p_putline(outfile, l) != l) return 3;
    }
    memcpy(linebuf(outfile), "\r\n", 2);
    if (p_putline(outfile, 2) != 2) return 3;
    p_close(outfile);
    return 0;
}

/* Pjcolor() takes an array of indices from the ditherline() routine and
** shuffles the bits around into the bizarre order the PaintJet needs.  Those
** of you without 16 fingers must be content that it works.
*/
static int pjcolor(U8 *indices)
{
    static char pcode2[20], pcode3[20];
    int w, p, x, cl, l;
    register U8 b1, b2;
    U8 *ip;
    U32 mem;
    U16 m0, m1, m2, m3, *pp, *patline;
    U8 plane[3][2][180];
    char *bp;

    mem = mark();
    if ((patline = (U16 *)talloc(pwidth * sizeof(U16))) == NULL) return 2;
    ip = indices;
    pp = patline;
	if (new->flags & 1) {
		ip += pwidth;
		for (x=0; x<pwidth; ++x) *pp++ = pjpatterns[*--ip];
	} else {
		for (x=0; x<pwidth; ++x) *pp++ = pjpatterns[*ip++];
	}

    w = (pwidth+3) >> 2;
    sprintf(pcode2, "\033*b%dV", w);
    sprintf(pcode3, "\033*b%dW", w);
    cl = strlen(pcode3);

    for (p=0; p<3; ++p) {
        m3 = 1 << p; m2 = m3 << 4; m1 = m2 << 4; m0 = m1 << 4;
        pp = patline;

        for (x=0; x<w; ++x) {
            b1 = b2 = 0;
            if (*pp   & m0) b1 |= 0x80;
            if (*pp   & m1) b1 |= 0x40;
            if (*pp   & m2) b2 |= 0x80;
            if (*pp++ & m3) b2 |= 0x40;

            if (*pp   & m0) b1 |= 0x20;
            if (*pp   & m1) b1 |= 0x10;
            if (*pp   & m2) b2 |= 0x20;
            if (*pp++ & m3) b2 |= 0x10;

            if (*pp   & m0) b1 |= 0x08;
            if (*pp   & m1) b1 |= 0x04;
            if (*pp   & m2) b2 |= 0x08;
            if (*pp++ & m3) b2 |= 0x04;

            if (*pp   & m0) b1 |= 0x02;
            if (*pp   & m1) b1 |= 0x01;
            if (*pp   & m2) b2 |= 0x02;
            if (*pp++ & m3) b2 |= 0x01;

            plane[p][0][x] = b1;
            plane[p][1][x] = b2;
        }
    }
    for (p=0; p<3; ++p) {
        bp = ((p==2) ? pcode3 : pcode2);
        l = strlen(bp);
        memcpy(linebuf(outfile), bp, l);
        if (p_putline(outfile, l) != l) return 3;
        memcpy(linebuf(outfile), plane[p][0], w);
        if (p_putline(outfile, w) != w) return 3;
    }
    for (p=0; p<3; ++p) {
        bp = ((p==2) ? pcode3 : pcode2);
        l = strlen(bp);
        memcpy(linebuf(outfile), bp, l);
        if (p_putline(outfile, l) != l) return 3;
        memcpy(linebuf(outfile), plane[p][1], w);
        if (p_putline(outfile, w) != w) return 3;
    }
    release(mem);
    return 0;
}

static int pjmono()
{
    pl_printf("Sorry, monochrome printing on the PaintJet has not been implemented\n");
	pl_printf("in version %s.  Issue COLOR command, the PRINT again.\n", version);
    return 0;
}

/* Print NEW buffer to HP PaintJet.
*/
static int pjprint()
{
    int i, r;
    U8 *rgb[3], *pjline;
    U16 lines;
    struct _plane *ip[3];

    if (new->planes != 3) return pjmono();
    if ((pwidth = new->width - xorigin) > 720) pwidth = 720;
    if ((r = startprint(printfile)) != 0) return r;

    startdither = 1;
    rv = pjrv; gv = pjgv; bv = pjbv; table2 = pjtable2;

    if ((pjline = (U8 *)talloc(pwidth)) == NULL) return 2;
    for (i=0; i<3; ++i) if ((ip[i] = openplane(i, new, READ)) == NULL) return 3;

    for (lines=0; lines < new->height; ++lines) {
        for (i=0; i<3; ++i) if (getline(ip[i]) == 0) return 4;
        for (i=0; i<3; ++i) rgb[i] = (U8 *)(ip[i]->linebuf + xorigin);
        ditherline(pwidth, rgb, pjline);
        if ((r = pjcolor(pjline)) != 0) return r;
        pl_trace(lines);
    }

    for (i=0; i<3; ++i) closeplane(ip[i]);
    if ((r = endprint()) != 0) return r;
    return 0;
}

/* Print NEW buffer to HP LaserJet.
*/
static int ljprint()
{
    int r, l;
    U8 *gray, *bitmap[2];
    U16 lines, bw;
    struct _plane *ip;
    char code[20];

    if (new->planes != 1) {
        pl_printf("Image must be grayscale to print on LaserJet.\n");
        return 0;
    }
    if (laserdpi == 300) bleed = 70; else bleed = 20;
    if ((pwidth = new->width - xorigin) > 1200) pwidth = 1200;
    if ((r = startprint(printfile)) != 0) return r;

    bw = (2 * pwidth + 7) >> 3;
    startdither = 1;
    bitmap[0] = (U8 *)talloc(bw);
    if ((bitmap[1] = (U8 *)talloc(bw)) == NULL) return 2;
    if ((ip = openplane(0, new, READ)) == NULL) return 3;

    for (lines=0; lines < new->height; ++lines) {
        if (getline(ip) == 0) return 4;
        gray = (U8 *)(ip->linebuf + xorigin);
        bitmapline(pwidth, gray, bitmap);
        sprintf(code, "\033*b%dW", bw);
        l = strlen(code);
        memcpy(linebuf(outfile), code, l);
        if (p_putline(outfile, l) != l) return 3;
        memcpy(linebuf(outfile), bitmap[0], bw);
        if (p_putline(outfile, bw) != bw) return 3;
        memcpy(linebuf(outfile), code, l);
        if (p_putline(outfile, l) != l) return 3;
        memcpy(linebuf(outfile), bitmap[0], bw);
        if (p_putline(outfile, bw) != bw) return 3;
        pl_trace(lines);
    }
    closeplane(ip);
    if ((r = endprint()) != 0) return r;
    return 0;
}

/* Command handler for PRINT.  Arguments, if any, are stored and printed
** after image.
*/
int printpic(int ac, argument *av)
{
    int r;
    char *pt;

    pt = printer;
    if (new->planes == 0) return 7;
    if (ac > 1) {
        pt = (*++av).cval;
        ntitles = 0;
        while (--ac > 1) titles[ntitles++] = (*++av).cval;
    }
    if ((ptype = pl_match(pt, printers)) == 0) {
        pl_printf("Unrecognized printer type.\n");
        return 0;
    } else {
        pl_printf("Printing...\n");
        if ((outbuf = (U8 *)talloc(BUFSIZE+512)) == NULL) return 2;
        switch (ptype) {
            case 1:     r = pjprint();      break;
            case 2:     r = ljprint();      break;
            default:    return 1;
        }
    }
    if (r == 0) pl_printf(done);
    return r;
}
